/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/*jshint browser: true *//* eslint-env browser */

/**
 * Connection Management for a browser.
 *
 * @module env/ConnectionManager
 * @private
 */
define([
  "bajaScript/comm",
  "bajaScript/env/HttpOnlyConnection",
  "bajaScript/env/WorkbenchConnection",
  "bajaScript/env/WebSocketConnection",
  "bajaScript/env/ReusedWebSocketConnection",
  "bajaPromises" ], function (
  baja,
  HttpOnlyConnection,
  WorkbenchConnection,
  WebSocketConnection,
  ReusedWebSocketConnection,
  bajaPromises) {

  "use strict";

  /**
   * Manages full duplex connections for the browser environment.
   *
   * @private
   * @class
   * @alias module:env/ConnectionManager
   * @param {Function} processNext A function called to process the next frame.
   */
  var ConnectionManager = function (processNext) {
    var that = this,
        sentFrames = that.$sentFrames = [];

    that.$connection = null;
    that.$handlers = {
      
     /**
      * A handler called when a connection is closed.
      *
      * @private
      * 
      * @param [ev] An optional event object.
      */
      close: function (ev) {
        var i;

        if (!baja.isStopping() && !baja.isStopped()) {
          // If there are any outstanding messages then fail them.
          for (i = 0; i < sentFrames.length; ++i) {
            sentFrames[i].cb.fail(ev || "Unknown");
          }
          sentFrames.splice(0, sentFrames.length);
        }

        // Remove BajaScript from event mode.
        baja.comm.stopEventMode();

        // Call to process the next frame.
        processNext();
      },

      /**
       * A handler called when a new message is received from a connection.
       *
       * @private
       * 
       * @param ev An event object.
       */
      message: function (ev) {
        // Parse the response message into a BOX frame
        var boxFrame = JSON.parse(ev.data),
            i,
            matchedFrame,
            frameData;
      
        // Bail if not a BOX frame
        if (!boxFrame || boxFrame.p !== "box") {
          baja.error("Invalid BOX frame: " + ev.data);
          return;
        }

        if (typeof boxFrame.n === "number") {
          // Find a match for the incoming frame.
          for (i = 0; i < sentFrames.length; ++i) {

            frameData = sentFrames[i];
            if (frameData.frame.getSequenceNum() === boxFrame.n) {
              try {                 
                // Make OK callback
                frameData.cb.ok(boxFrame);              
              } finally {
                matchedFrame = true;
                sentFrames.splice(i, 1);

                if (frameData.processed) {
                  frameData.processed();
                }

                break;
              }
            }

          }
        }

        if (!matchedFrame) {
          var messages = boxFrame.m;
          // Process unsolicited BOX frames
          for (i = 0; i < messages.length; ++i) {
            var message = messages[i];
            // Handle unsolicited events
            if (message.t === "u") {
              var body = message.b;
              if (validForMyServerSession(body)) {
                baja.comm.handleEvents(body.d);
              }
            }
          }
        }
      },

      /**
       * A handler called when a connection is opened.
       *
       * @private
       */
      open: function () {
        // Put BajaScript into event mode.
        baja.comm.startEventMode();
      },

      /**
       * A handler called when a connection goes into error.
       *
       * @private
       * 
       * @param [ev] An optional error object.
       */
      error: function (ev) {
        baja.error(ev);
      }
    };
  };

  function validForMyServerSession(messageBody) {
    var id = messageBody.id;
    return !id || (id === baja.comm.getServerSessionId());
  }
  
  ConnectionManager.prototype.$makeConnection = function (params) {
    var handlers = this.$handlers,
      hasWebSockets = typeof WebSocket === 'function';

    // eslint-disable-next-line camelcase
    if (typeof niagara_wb_util_sendBox === 'function') {
      return new WorkbenchConnection(handlers);
    } else if (hasWebSockets && params) {
      return new ReusedWebSocketConnection(handlers, params);
    } else if (hasWebSockets) {
      return new WebSocketConnection(handlers);
    }

    return new HttpOnlyConnection();
  };

  /**
   * Starts the connection or short circuits if a connection is already available.
   *
   * @private
   * 
   * @param [params] Optional parameters used to create the underlying connection.
   * @returns {Promise} promise to be resolved when the connection is
   * authenticated and opened.
   */
  ConnectionManager.prototype.start = function (params) {
    var that = this;

    if (that.$connection) {
      return bajaPromises.resolve();
    }
  
    return that.$doAuthenticate(params)
      .then(function () { return that.$doStart(); });
  };

  /**
   * Instantiates the connection and uses it to authenticate with the server.
   *
   * @private
   * @param {object} [params] params to be used to instantiate the connection
   * @returns {Promise}
   */
  ConnectionManager.prototype.$doAuthenticate = function (params) {
    var connection = this.$connection = this.$makeConnection(params);

    return bajaPromises.resolve()
      .then(function () {
        return connection.authenticate();
      });
  };

  /**
   * Starts up the authenticated connection. If the connection fails to start,
   * fall back to HTTP-only (no web sockets or interop) and try again (without
   * re-authenticating).
   * @private
   * @returns {Promise}
   */
  ConnectionManager.prototype.$doStart = function () {
    var that = this,
      connection = that.$connection;
    return bajaPromises.resolve()
      .then(function () {
        return connection.start();
      })
      .catch(function () {
        connection = that.$connection = new HttpOnlyConnection();
        return connection.start();
      });
  };

  /**
   * Send the data to the connection.
   *
   * @private
   * 
   * @param frameData The frame data to send.
   * @returns {Boolean} returns true if successfully sent or false if it can't be sent.
   * If false is returned then the message can't be sent. It may be resent by another medium
   * (i.e. synchronously via HTTP or some other medium).
   */
  ConnectionManager.prototype.send = function (frameData) {
    var conn = this.$getOpenConnection();
    // If WebSockets aren't support or the frame
    // needs to be sent synchronously or the web socket isn't open yet
    // then send via HTTP.
    if (frameData.frame.$sync || !conn) {
      return false;
    }

    // Remember the BOX frame we've sent so we can reference it on the response.
    this.$sentFrames.push(frameData);
                                 
    // Send the JSON message through the WebSocket
    conn.send(frameData);
           
    // Return true so we can send it
    return true;
  };

  /**
   * Send the data as a beacon to the connection, if the connection implements
   * `sendBeacon`.
   *
   * @private
   *
   * @param frameData The frame data to send.
   * @returns {Boolean} returns true if successfully sent or false if it can't
   * be sent. If false is returned then the message can't be sent. It may be
   * resent by another medium (i.e. synchronously via HTTP or some other
   * medium).
   */
  ConnectionManager.prototype.sendBeacon = function (frameData) {
    var frame = frameData.frame,
      syncFlag = frame.$sync,
      conn = this.$getOpenConnection();

    if (syncFlag !== 'beacon' || !conn || typeof conn.sendBeacon !== 'function') {
      return false;
    }

    conn.sendBeacon(frameData);

    return true;
  };

  /**
   * @private
   * @returns {boolean}
   */
  ConnectionManager.prototype.$getOpenConnection = function () {
    var conn = this.$connection;
    return conn && conn.isOpen() && conn;
  };

  /**
   * Close any underlying connections.
   *
   * @private
   */
  ConnectionManager.prototype.close = function () {
    if (this.$connection) {
      this.$connection.close();
    }
  };

  /**
   * Returns the underlying internal connection object or null if it doesn't exist.
   *
   * @private
   * 
   * @returns The internal connection or null if there isn't one.
   */
  ConnectionManager.prototype.getInternalConnection = function () {
    return this.$connection && this.$connection.getInternalConnection();
  };

  return ConnectionManager;
});
